#include <fstream>

#include <wx/file.h>
#include <wx/textfile.h>
#include <wx/wfstream.h>
#include <wx/txtstrm.h>

#include "resultlist.h"
#include "guiApp.h"

BEGIN_EVENT_TABLE(ResultList, wxGrid)
    EVT_GRID_SELECT_CELL(ResultList::OnResultListClick)
	EVT_GRID_CELL_CHANGE(ResultList::OnResultListChange)
	EVT_COMMAND(wxID_ANY, EVT_CREATE_LIST, ResultList::OnCreateList)
END_EVENT_TABLE()

void GetWeightMinMax(Graph *g, float& weightMin, float& weightMax) {
	weightMin = FLT_MAX;
	weightMax = FLT_MIN;

	float weight;
	Graph::neighbor_it n_iter;
	for (std::size_t i = 0; i < g->num_vertices(); i++) {
		for (n_iter = g->neighbors_begin(i); n_iter != g->neighbors_end(i); n_iter++) {
			if (i > n_iter->neighbor) continue;
			weight = g->edge_weight(i, n_iter->neighbor);
			weightMin = std::min(weightMin, weight);
			weightMax = std::max(weightMax, weight);
		}
	}
}

ResultList::ResultList(wxWindow* parent, wxWindowID id, const wxPoint& pos, 
				const wxSize& size)	: wxGrid(parent, id, pos, size) {
	sel_old = -1;
	SaveFileDialog =  new wxFileDialog(this, wxT("Save result list to file"), wxT(""), wxT(""), 
		wxT("ResultList (*.rsl)|*.rsl|Character Separated Values (*.csv)|*.csv|All files (*)|*"), wxSAVE);
	OpenFileDialog =  new wxFileDialog(this, wxT("Load result list from file"), wxT(""), wxT(""), 
		wxT("ResultList (*.rsl)|*.rsl|All files (*)|*"), wxOPEN);
}

ResultList::~ResultList() {
	delete SaveFileDialog;
	delete OpenFileDialog;
}

void ResultList::OnResultListClick(wxGridEvent& event) {
	int sel = event.GetRow();
	if (sel == sel_old) {event.Skip(); return; };

	sel_old = sel;
	DisplayPath(sel);
	event.Skip();
}


void ResultList::OnResultListChange(wxGridEvent& event) {
	int sel = event.GetRow();
	int seltab = wxGetApp().frame->ntb_Graphs->GetSelection();
	if (seltab == -1) {event.Skip(); return; }

	GraphDiagram* diag = (GraphDiagram*)(wxGetApp().frame->Canvas[seltab]->GetDiagram());

	PathSet::it iter = paths->begin();
	for(int i = 0; i < sel; i++) iter++;
	std::size_t path_length = iter->path().size();
	
	if ((wxAtoi(GetCellValue(sel, path_length + 1)) == 1) && (event.GetCol() != path_length + 1)) {
			std::size_t close_dist = wxAtoi(GetCellValue(sel, path_length + 2));
			std::size_t far_dist = wxAtoi(GetCellValue(sel, path_length + 3));
			float min_prob = wxAtof(GetCellValue(sel, path_length + 4));
			if (!resModIsSPD) min_prob = exp(-min_prob);
			diag->GetDiagramContent()->ModPath(iter->path(), PathParam(close_dist, far_dist, min_prob));
	}

	if (event.GetCol() == path_length + 1) {
		if (wxAtoi(GetCellValue(sel, path_length + 1)) == 1) {
			std::size_t close_dist = wxAtoi(GetCellValue(sel, path_length + 2));
			std::size_t far_dist = wxAtoi(GetCellValue(sel, path_length + 3));
			float min_prob = wxAtof(GetCellValue(sel, path_length + 4));
			if (!resModIsSPD) min_prob = exp(-min_prob);
			diag->GetDiagramContent()->AddPath(iter->path(), PathParam(close_dist, far_dist, min_prob));
		}
		else{
			diag->GetDiagramContent()->DelPath(iter->path());
		}
	}
	DisplayPath(sel);
	event.Skip();
}

void ResultList::DisplayPath(std::size_t path_nr) {
	int sel = wxGetApp().frame->ntb_Graphs->GetSelection();
	if (sel == -1) return;

	GraphCanvas* gcan = wxGetApp().frame->Canvas[sel];
	GraphDiagram* diag = (GraphDiagram*)(gcan->GetDiagram());

	diag->modIsSPD = resModIsSPD;
	diag->weightMin = this->weightMin;
	diag->weightMax = this->weightMax;
	diag->weightDiff = this->weightMax - this->weightMin;

	PathSet::it iter = paths->begin();
	for(int i = 0; i < path_nr; i++) iter++;

	std::size_t path_length = iter->path().size();
	diag->DeleteAllShapes();
	diag->xPos.clear();
	diag->yPos.clear();

	std::size_t close_dist = wxAtoi(GetCellValue(path_nr, path_length + 2));
	std::size_t far_dist = wxAtoi(GetCellValue(path_nr, path_length + 3));
	wxString prob = GetCellValue(path_nr, path_length + 4);
	float min_prob = wxAtof(prob);
	if (!resModIsSPD) min_prob = exp(-min_prob);

	Agraph_t* agraph = diag->GetDiagramContent()->ComputeLayout(*g, (iter->path()), PathParam(close_dist, far_dist, min_prob));
	FILE *dotfile = fopen("./currentgraph.dot", "w");
	agwrite (agraph, dotfile);
	fclose(dotfile);
	gcan->DrawGraph(agraph, g);
	agclose(agraph);

	wxGetApp().frame->ShowLegend(true);
}

void ResultList::UpdateParam(Content* cont) {
	if (GetNumberRows() == 0) return;

	PathSet::it iter = paths->begin();
	std::size_t path_length = iter->path().size();
	for (int i = 0; i < paths->size(); i++) {
		PathParam* param = cont->GetPathParam(iter->path());
		if (param == NULL) {
			SetCellValue(i, path_length + 1, wxT("0"));
		} else {
			wxString close_dist;
			wxString far_dist;
			wxString min_prob;
			close_dist << param->close_dist;
			far_dist << param->far_dist;
			min_prob << param->min_prob;
			SetCellValue(i, path_length + 1, wxT("1"));
			SetCellValue(i, path_length + 2, close_dist);
			SetCellValue(i, path_length + 3, far_dist);
			SetCellValue(i, path_length + 4, min_prob);
		}
		iter++;
	}	
}

void ResultList::SaveList() {
    if (SaveFileDialog->ShowModal() == wxID_OK) 
	{		
		wxString filename = SaveFileDialog->GetPath();
		if (SaveFileDialog->GetFilterIndex() == 0 && (filename.Right(4).Cmp(wxT(".rsl")) != 0)) filename.Append(wxT(".rsl"));
		if (SaveFileDialog->GetFilterIndex() == 1 && (filename.Right(4).Cmp(wxT(".csv")) != 0)) filename.Append(wxT(".csv"));
		
		bool csv = false;
		wxString sepChar = wxT(" ");
		if (SaveFileDialog->GetFilterIndex() == 1) {
			csv = true;
			sepChar = wxT(";");
		}

		wxFile file;
		std::size_t path_nr = 0, j;
		wxString line;

		file.Create(filename, true, 0755);
		if (!csv) {	
			line << (resModIsSPD ? 0 : 1);
			file.Write(line);
			file.Write(wxTextFile::GetEOL());
			file.Write(graph_filename);
			file.Write(wxTextFile::GetEOL());
			line.Clear();
			line << paths->size() << wxT(" ") << paths->begin()->path().size();
			file.Write(line);
			file.Write(wxTextFile::GetEOL());
		}

    	for (PathSet::it iter = paths->begin(); iter != paths->end(); ++iter) {
			line.Clear();
        	line << ((!csv || !resModIsSPD)? iter->path_weight() : exp(-iter->path_weight())); 
	    	for (j = 0; j < iter->path().size(); ++j) 
				if (!csv) line << sepChar << iter->path()[j];
				else line << sepChar << wxString(g->vertex_name(iter->path()[j]).c_str(), wxConvUTF8);

			if (!csv) {	
				line << sepChar
					<< GetCellValue(path_nr, j + 1) << sepChar 
					<< GetCellValue(path_nr, j + 2) << sepChar
					<< GetCellValue(path_nr, j + 3) << sepChar
					<< GetCellValue(path_nr, j + 4);
			}
			file.Write(line);
			file.Write(wxTextFile::GetEOL());
			path_nr++;
		}
		file.Close();
	}
}

bool ResultList::LoadList() {
    if (OpenFileDialog->ShowModal() == wxID_OK) {
		wxString filename = OpenFileDialog->GetPath();
		wxFileInputStream fileinstream(filename);
		wxTextInputStream instream(fileinstream);

		vertex_t v;
		std::size_t num_paths, path_length, selected, close_dist, far_dist;
		weight_t weight;
		float min_prob;
		int mod;		
		std::vector<std::size_t> selections;
		std::vector<PathParam> params;

		instream >> mod;
		if (mod == 0) resModIsSPD = true;
		else resModIsSPD = false;

		graph_filename = instream.ReadLine();
		instream >> num_paths >> path_length;
		if ((num_paths <= 0) || (path_length <= 1)) {
			wxMessageBox(wxT("Error while loading file!\nFormat incorrect.\nLoading aborted."), wxT("Error"), wxICON_WARNING);
			return false;
		}
		paths = new PathSet(num_paths, 100);

		std::size_t lineno = 0;
		while (!fileinstream.Eof()) {
			instream >> weight;
			std::vector<vertex_t> path;
			for (std::size_t j = 0; j < path_length; j++) {
				instream >> v;
				path.push_back(v);
			}
			instream >> selected >> close_dist >> far_dist >> min_prob;
			if (fileinstream.LastRead() == 0) break;
			paths->add(path, weight);
			selections.push_back(selected);
			params.push_back(PathParam(close_dist, far_dist, min_prob));
			lineno++;
		}
		if (lineno != num_paths) {
			wxMessageBox( wxT("Error while loading file!\nNumber of data sets incorrect.\nLoading aborted."), wxT("Error"), wxICON_ERROR);
			return false;
		}
		if (!wxFile::Exists(graph_filename)) {
			wxString msg;
			msg << wxT("This result list corresponds to the graph file\n") << graph_filename
				<< wxT("\nwhich does not exist anymore.\nPlease specify the new location!");
			wxMessageBox(msg, wxT("Warning"), wxICON_WARNING);
		    if (wxGetApp().frame->OpenFileDialog->ShowModal() == wxID_OK) 
				graph_filename = wxGetApp().frame->OpenFileDialog->GetPath();
			else return false;
		}
		std::ifstream graphin(((std::string)graph_filename.mb_str(wxConvUTF8)).c_str());
		wxGetApp().graphload_ok = true;
		g = new Graph(graphin, !resModIsSPD);
		if (!(wxGetApp().graphload_ok)) return false;
		if (g->num_vertices() == 0) {
			wxMessageBox(wxT("Error while loading graph!\nGraph is empty.\nLoading aborted."), wxT("Error"), wxICON_ERROR);
			return false;
		}
		wxGetApp().frame->AddGraphTab(wxGetApp().frame->ntb_Graphs);
		CreateList();
		SetSelections(selections);
		SetParams(params);
		return true;
	}
	return false;
}

void ResultList::CreateList() {
	GetWeightMinMax(g, weightMin, weightMax);
	std::size_t path_length = paths->begin()->path().size();

	CreateGrid( paths->size(), path_length + 5, wxGrid::wxGridSelectRows);
	SetRowLabelSize(40);
	SetColLabelSize(30);
	EnableScrolling(TRUE, TRUE);

	int x, y;	
	this->GetVirtualSize(&x, &y);
	this->SetVirtualSize(x, y);

	this->SetColLabelValue(0, (resModIsSPD ? wxT("Prob") : wxT("Weight")));
	for (int i = 1; i <= path_length; i++) {
		wxString label;
		label << wxT("Prot ") << i;
		this->SetColLabelValue(i, label);
	}

	this->SetColLabelValue(path_length + 1, wxT("Selected"));
	this->SetColLabelValue(path_length + 2, wxT("MaxDist Con"));
	this->SetColLabelValue(path_length + 3, wxT("MaxDist Uncon"));
	this->SetColLabelValue(path_length + 4, (resModIsSPD ? wxT("Min Prob") : wxT("Max Weight")));

    size_t i2 = 0;
    for (PathSet::it i = this->paths->begin(); i != this->paths->end(); ++i) {
	    wxString line;
        line << (resModIsSPD ? exp(-i->path_weight()) : i->path_weight());
		this->SetCellValue(i2, 0, line);
		this->SetReadOnly(i2, 0);
		std::size_t j;
	    for (j = 0; j < i->path().size(); ++j) {
			this->SetCellValue(i2, j + 1, wxString(g->vertex_name(i->path()[j]).c_str(), wxConvUTF8));
			this->SetReadOnly(i2, j + 1);
		}
		
		this->SetCellValue(i2, j + 1, _T("0"));
    	this->SetCellRenderer(i2, j + 1, new wxGridCellBoolRenderer);
    	this->SetCellEditor(i2, j + 1, new wxGridCellBoolEditor);

		const wxString choices[] = { _T("0"), _T("1"), _T("2"), _T("3"), _T("4"), _T("5")};
//	  	this->SetCellEditor(i2, j + 2, new wxGridCellChoiceEditor(WXSIZEOF(choices), choices));
		this->SetCellValue(i2, j + 2, choices[0]);
//		this->SetCellOverflow(i2, j + 2, false);

//		this->SetCellEditor(i2, j + 3, new wxGridCellChoiceEditor(WXSIZEOF(choices), choices));
		this->SetCellValue(i2, j + 3, choices[0]);
//		this->SetCellOverflow(i2, j + 3, false);

		wxString min_prob;
		min_prob << (resModIsSPD ? 0.7 : 0.15);
		this->SetCellValue(i2, j + 4, min_prob);
	    i2++;
    }
	wxGetApp().Yield();
	if (paths->size() > 0) {
		SelectRow(0);
		sel_old = 0;
		DisplayPath(0);
		int sel = wxGetApp().frame->ntb_Graphs->GetSelection();
		if (sel > -1) {
			wxGetApp().frame->Canvas[sel]->ZoomFit();
			if (wxGetApp().frame->Diagram[sel]->g_scale < 0.4) {
				wxGetApp().frame->Canvas[sel]->Zoom(0.4);
				wxGetApp().frame->cbx_Zoom->SetSelection(3);
			}		
		}
	}
}

void ResultList::OnCreateList(wxCommandEvent& WXUNUSED(event)) {
     CreateList();     
}

void ResultList::SetSelections(const std::vector<std::size_t>& selections) {
	std::vector<std::size_t>::const_iterator iter;
	std::size_t row_nr = 0;
	std::size_t path_length = paths->begin()->path().size();
	for(iter = selections.begin(); iter != selections.end(); iter++) {
		wxString value;
		value << *iter;
		SetCellValue(row_nr, path_length + 1, value);
		row_nr++;
	}
}

void ResultList::SetParams(const std::vector<PathParam>& params) {
	std::vector<PathParam>::const_iterator iter;
	std::size_t row_nr = 0;
	std::size_t path_length = paths->begin()->path().size();
	for(iter = params.begin(); iter != params.end(); iter++) {
		wxString close_dist;
		wxString far_dist;
		wxString min_prob;
		close_dist << iter->close_dist;
		far_dist << iter->far_dist;
		min_prob << iter->min_prob;
		SetCellValue(row_nr, path_length + 2, close_dist);
		SetCellValue(row_nr, path_length + 3, far_dist);
		SetCellValue(row_nr, path_length + 4, min_prob);
		row_nr++;
	}	
}
